@@ -19,9 +19,10 @@ import colorama
 import jsonschema
 import yaml
 
-from . import util
+from . import util as _util
 
 verbosity = False
+debug_manifests = False
 
 global_yaml_loader = yaml.SafeLoader
 global_bootstrap_validator = None
@@ -319,6 +320,12 @@ class Config:
 			return False
 		return self._site_yml['pkg_management']['format'] == 'xbps'
 
+	@property
+	def container_runtime(self):
+		if 'container' not in self._site_yml:
+			return None
+		return self._site_yml['container'].get('runtime')
+
 	@property
 	def source_root(self):
 		return os.path.join(os.getcwd(),
@@ -328,6 +335,14 @@ class Config:
 	def build_root(self):
 		return os.getcwd()
 
+	@property
+	def sysroot_subdir(self):
+		if 'directories' not in self._root_yml or 'system_root' not in self._root_yml['directories']:
+			return 'system-root'
+		else:
+			return self._root_yml['directories']['system_root']
+
+	# sysroot_dir = build_root + sysroot_subdir
 	@property
 	def sysroot_dir(self):
 		if 'directories' not in self._root_yml or 'system_root' not in self._root_yml['directories']:
@@ -339,6 +354,14 @@ class Config:
 	def xbps_repository_dir(self):
 		return os.path.join(self.build_root, 'xbps-repo')
 
+	@property
+	def tool_build_subdir(self):
+		if 'directories' not in self._root_yml or 'pkg_builds' not in self._root_yml['directories']:
+			return 'tool-builds'
+		else:
+			return self._root_yml['directories']['tool_builds']
+
+	# tool_build_dir = build_root + tool_build_subdir.
 	@property
 	def tool_build_dir(self):
 		if 'directories' not in self._root_yml or 'pkg_builds' not in self._root_yml['directories']:
@@ -346,6 +369,14 @@ class Config:
 		else:
 			return os.path.join(self.build_root, self._root_yml['directories']['tool_builds'])
 
+	@property
+	def pkg_build_subdir(self):
+		if 'directories' not in self._root_yml or 'pkg_builds' not in self._root_yml['directories']:
+			return 'pkg-builds'
+		else:
+			return self._root_yml['directories']['pkg_builds']
+
+	# pkg_build_dir = build_root + pkg_build_subdir.
 	@property
 	def pkg_build_dir(self):
 		if 'directories' not in self._root_yml or 'pkg_builds' not in self._root_yml['directories']:
@@ -353,6 +384,14 @@ class Config:
 		else:
 			return os.path.join(self.build_root, self._root_yml['directories']['pkg_builds'])
 
+	@property
+	def tool_out_subdir(self):
+		if 'directories' not in self._root_yml or 'tools' not in self._root_yml['directories']:
+			return 'tools'
+		else:
+			return self._root_yml['directories']['tools']
+
+	# tool_out_dir = build_root + tool_out_subdir
 	@property
 	def tool_out_dir(self):
 		if 'directories' not in self._root_yml or 'tools' not in self._root_yml['directories']:
@@ -360,6 +399,14 @@ class Config:
 		else:
 			return os.path.join(self.build_root, self._root_yml['directories']['tools'])
 
+	@property
+	def package_out_subdir(self):
+		if 'directories' not in self._root_yml or 'packages' not in self._root_yml['directories']:
+			return 'packages'
+		else:
+			return self._root_yml['directories']['packages']
+
+	# package_out_dir = build_root + package_out_subdir
 	@property
 	def package_out_dir(self):
 		if 'directories' not in self._root_yml or 'packages' not in self._root_yml['directories']:
@@ -367,6 +414,11 @@ class Config:
 		else:
 			return os.path.join(self.build_root, self._root_yml['directories']['packages'])
 
+	@property
+	def all_options(self):
+		for yml in self._root_yml.get('declare_options', []):
+			yield yml['name']
+
 	def get_option_value(self, name):
 		decl = None
 		for yml in self._root_yml.get('declare_options', []):
@@ -424,6 +476,10 @@ class ScriptStep:
 			return None
 		return self._step_yml['workdir']
 
+	@property
+	def containerless(self):
+		return self._step_yml.get('containerless', False)
+
 	@property
 	def quiet(self):
 		if 'quiet' not in self._step_yml:
@@ -646,6 +702,13 @@ class Source(RequirementsMixin):
 			return os.path.join(self._cfg.source_root, self._this_yml['subdir'])
 		return self._cfg.source_root
 
+	@property
+	def source_subdir(self):
+		if 'subdir' in self._this_yml:
+			return os.path.join(self._this_yml['subdir'], self._name)
+		return self._name
+
+	# source_dir = source_root + source_subdir.
 	@property
 	def source_dir(self):
 		return os.path.join(self.sub_dir, self._name)
@@ -919,10 +982,18 @@ class HostPackage(RequirementsMixin):
 					continue
 				yield yml['tool']
 
+	@property
+	def build_subdir(self):
+		return os.path.join(self._cfg.tool_build_subdir, self.name)
+
 	@property
 	def build_dir(self):
 		return os.path.join(self._cfg.tool_build_dir, self.name)
 
+	@property
+	def prefix_subdir(self):
+		return os.path.join(self._cfg.tool_out_subdir, self.name)
+
 	@property
 	def prefix_dir(self):
 		return os.path.join(self._cfg.tool_out_dir, self.name)
@@ -1043,6 +1114,10 @@ class TargetPackage(RequirementsMixin):
 			return self._this_yml['source']['name']
 		return self.name
 
+	@property
+	def build_subdir(self):
+		return os.path.join(self._cfg.pkg_build_subdir, self.name)
+
 	@property
 	def build_dir(self):
 		return os.path.join(self._cfg.pkg_build_dir, self.name)
@@ -1051,6 +1126,10 @@ class TargetPackage(RequirementsMixin):
 	def staging_dir(self):
 		return os.path.join(self._cfg.package_out_dir, self.name)
 
+	@property
+	def collect_subdir(self):
+		return os.path.join(self._cfg.package_out_subdir, self.name + '.collect')
+
 	@property
 	def collect_dir(self):
 		return os.path.join(self._cfg.package_out_dir, self.name + '.collect')
@@ -1263,83 +1342,55 @@ class RunTask(RequirementsMixin):
 def config_for_dir():
 	return Config('')
 
-class EnvironmentComposer:
-	def __init__(self, cfg):
-		self.cfg = cfg
-		self.path_dirs = [ ]
-		self.shared_lib_dirs = [ ]
-		self.aclocal_dirs = [ ]
+def execute_manifest(manifest):
+	source_root = manifest['source_root']
+	build_root = manifest['build_root']
+	sysroot_dir = os.path.join(manifest['build_root'], manifest['sysroot_subdir'])
 
-	def compose(self, for_package=False, explicit_pkgconfig=False):
-		environ = os.environ.copy()
+	def substitute(varname):
+		if varname == 'SOURCE_ROOT':
+			return source_root
+		elif varname == 'BUILD_ROOT':
+			return build_root
+		elif varname == 'SYSROOT_DIR':
+			return sysroot_dir
+		elif varname == 'PARALLELISM':
+			nthreads = get_concurrency()
+			return str(nthreads)
+		elif varname.startswith('OPTION:'):
+			return manifest['option_values'][varname[7:]]
 
-		environ['XBSTRAP_SOURCE_ROOT'] = self.cfg.source_root
-		environ['XBSTRAP_BUILD_ROOT'] = self.cfg.build_root
-		environ['XBSTRAP_SYSROOT_DIR'] = self.cfg.sysroot_dir
-
-		if for_package and not explicit_pkgconfig:
-			pkgcfg_libdir = os.path.join(self.cfg.sysroot_dir, 'usr', 'lib', 'pkgconfig')
-			pkgcfg_libdir += ':' + os.path.join(self.cfg.sysroot_dir, 'usr', 'share', 'pkgconfig')
-
-			environ.pop('PKG_CONFIG_PATH', None)
-			environ['PKG_CONFIG_SYSROOT_DIR'] = self.cfg.sysroot_dir
-			environ['PKG_CONFIG_LIBDIR'] = pkgcfg_libdir
-
-		self._prepend_dirs(environ, 'PATH', self.path_dirs)
-		self._prepend_dirs(environ, 'LD_LIBRARY_PATH', self.shared_lib_dirs)
-		self._prepend_dirs(environ, 'ACLOCAL_PATH', self.aclocal_dirs)
-		return environ
-
-	def _prepend_dirs(self, environ, varname, dirs):
-		if not dirs:
-			return
-		joined = ':'.join(dirs)
-		if varname in environ and environ[varname]:
-			environ[varname] = joined + ':' + environ[varname]
-		else:
-			environ[varname] = joined
-
-	def _append_dirs(self, environ, varname, dirs):
-		if not dirs:
-			return
-		joined = ':'.join(dirs)
-		if varname in environ and environ[varname]:
-			environ[varname] += ':' + joined
-		else:
-			environ[varname] = joined
-
-def run_tool(cfg, args, tool_pkgs=[], virtual_tools=[], workdir=None, extra_environ=dict(),
-		for_package=False, quiet=False):
-	pkg_queue = []
-	pkg_visited = set()
-
-	for pkg in tool_pkgs:
-		assert pkg.name not in pkg_visited
-		pkg_queue.append(pkg)
-		pkg_visited.add(pkg.name)
-
-	i = 0 # Need index-based loop as pkg_queue is mutated in the loop.
-	while i < len(pkg_queue):
-		pkg = pkg_queue[i]
-		for dep_name in pkg.recursive_tools_required:
-			if dep_name in pkg_visited:
-				continue
-			dep_pkg = cfg.get_tool_pkg(dep_name)
-			pkg_queue.append(dep_pkg)
-			pkg_visited.add(dep_name)
-		i += 1
+		if manifest['context'] == 'source':
+			if varname == 'THIS_SOURCE_DIR':
+				return os.path.join(source_root, manifest['subject']['source_subdir'])
+		elif manifest['context'] == 'tool' or manifest['context'] == 'tool-stage':
+			if varname == 'THIS_SOURCE_DIR':
+				return os.path.join(source_root, manifest['subject']['source_subdir'])
+			elif varname == 'THIS_BUILD_DIR':
+				return os.path.join(build_root, manifest['subject']['build_subdir'])
+			elif varname == 'PREFIX':
+				return os.path.join(build_root, manifest['subject']['prefix_subdir'])
+		elif manifest['context'] == 'pkg':
+			if varname == 'THIS_SOURCE_DIR':
+				return os.path.join(source_root, manifest['subject']['source_subdir'])
+			elif varname == 'THIS_BUILD_DIR':
+				return os.path.join(build_root, manifest['subject']['build_subdir'])
+			elif varname == 'THIS_COLLECT_DIR':
+				return os.path.join(build_root, manifest['subject']['collect_subdir'])
 
 	# /bin directory for virtual tools.
 	explicit_pkgconfig = False
 	vb = tempfile.TemporaryDirectory()
 
-	for yml in virtual_tools:
+	for yml in manifest['virtual_tools']:
 		if yml['virtual'] == 'pkgconfig-for-host':
 			vscript = os.path.join(vb.name, yml['program_name'])
 			paths = []
-			for tool in pkg_queue:
-				paths.append(tool.prefix_dir + '/lib/pkgconfig')
-				paths.append(tool.prefix_dir + '/share/pkgconfig')
+			for tool_yml in manifest['tools']:
+				paths.append(os.path.join(build_root, tool_yml['prefix_subdir'],
+						'lib/pkgconfig'))
+				paths.append(os.path.join(build_root, tool_yml['prefix_subdir'],
+						'share/pkgconfig'))
 			with open(vscript, 'wt') as f:
 				f.write('#!/bin/sh\n'
 					+ 'PKG_CONFIG_PATH=' + ':'.join(paths)
@@ -1360,49 +1411,236 @@ def run_tool(cfg, args, tool_pkgs=[], virtual_tools=[], workdir=None, extra_envi
 		else:
 			raise RuntimeError("Unknown virtual tool {}".format(yml['virtual']))
 
-	composer = EnvironmentComposer(cfg)
-	composer.path_dirs.append(vb.name)
-	for pkg in pkg_queue:
-		composer.path_dirs.append(os.path.join(pkg.prefix_dir, 'bin'))
-		if pkg.exports_shared_libs:
-			composer.shared_lib_dirs.append(os.path.join(pkg.prefix_dir, 'lib'))
-		if pkg.exports_aclocal:
-			composer.aclocal_dirs.append(os.path.join(pkg.prefix_dir, 'share/aclocal'))
-	environ = composer.compose(for_package=for_package, explicit_pkgconfig=explicit_pkgconfig)
-	environ.update(extra_environ)
-
-	output = None # Default: Do not redirect output.
-	if quiet:
-		output = subprocess.DEVNULL
-
-	print("{}xbstrap{}: Running {} (tools: {})".format(
-			colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
-			args, [tool.name for tool in pkg_queue]))
-	subprocess.check_call(args, env=environ, cwd=workdir,
-			stdout=output, stderr=output)
-
-def run_step(cfg, step, default_workdir, substitute, tool_pkgs, virtual_tools,
-		for_package=False):
-	if isinstance(step.args, list):
-		args = [replace_at_vars(arg, substitute) for arg in step.args]
+	# Determine the arguments.
+	if isinstance(manifest['args'], list):
+		args = [replace_at_vars(arg, substitute) for arg in manifest['args']]
 	else:
-		assert isinstance(step.args, str)
-		args = ['/bin/bash', '-c', replace_at_vars(step.args, substitute)]
+		assert isinstance(manifest['args'], str)
+		args = ['/bin/bash', '-c', replace_at_vars(manifest['args'], substitute)]
 
-	environ = dict()
-	for (key, value) in step.environ.items():
+	# Build the environment
+	environ = os.environ.copy()
+
+	path_dirs = [vb.name]
+	ldso_dirs = []
+	aclocal_dirs = []
+	for yml in manifest['tools']:
+		prefix_dir = os.path.join(build_root, yml['prefix_subdir'])
+		path_dirs.append(os.path.join(prefix_dir, 'bin'))
+		if yml['exports_shared_libs']:
+			ldso_dirs.append(os.path.join(prefix_dir, 'lib'))
+		if yml['exports_aclocal']:
+			aclocal_dirs.append(os.path.join(prefix_dir, 'share/aclocal'))
+
+	_util.build_environ_paths(environ, 'PATH', prepend=path_dirs)
+	_util.build_environ_paths(environ, 'LD_LIBRARY_PATH', prepend=ldso_dirs)
+	_util.build_environ_paths(environ, 'ACLOCAL_PATH', prepend=aclocal_dirs)
+
+	if manifest['for_package'] and not explicit_pkgconfig:
+		pkgcfg_libdir = os.path.join(sysroot_dir, 'usr', 'lib', 'pkgconfig')
+		pkgcfg_libdir += ':' + os.path.join(sysroot_dir, 'usr', 'share', 'pkgconfig')
+
+		environ.pop('PKG_CONFIG_PATH', None)
+		environ['PKG_CONFIG_SYSROOT_DIR'] = sysroot_dir
+		environ['PKG_CONFIG_LIBDIR'] = pkgcfg_libdir
+
+	environ['XBSTRAP_SOURCE_ROOT'] = source_root
+	environ['XBSTRAP_BUILD_ROOT'] = build_root
+	environ['XBSTRAP_SYSROOT_DIR'] = sysroot_dir
+
+	for key, value in manifest['extra_environ'].items():
 		environ[key] = replace_at_vars(value, substitute)
 
-	if step.workdir is not None:
-		workdir = replace_at_vars(step.workdir, substitute)
+	# Determine the stdout sink.
+	output = None # Default: Do not redirect output.
+	if manifest['quiet']:
+		output = subprocess.DEVNULL
+
+	# Determine the working directory.
+	if manifest['workdir'] is not None:
+		workdir = replace_at_vars(manifest['workdir'], substitute)
 	else:
-		workdir = default_workdir
+		if manifest['context'] == 'source':
+			workdir = os.path.join(source_root, manifest['subject']['source_subdir'])
+		elif manifest['context'] == 'tool' or manifest['context'] == 'tool-stage':
+			workdir = os.path.join(build_root, manifest['subject']['build_subdir'])
+		elif manifest['context'] == 'pkg':
+			workdir = os.path.join(build_root, manifest['subject']['build_subdir'])
+		elif manifest['context'] == 'task':
+			workdir = source_root
+		elif manifest['context'] == 'tool-task':
+			workdir = os.path.join(build_root, manifest['subject']['build_subdir'])
+		elif manifest['context'] == 'pkg-task':
+			workdir = os.path.join(build_root, manifest['subject']['build_subdir'])
+		else:
+			raise RuntimeError("Unexpected context")
 
-	quiet = step.quiet and not verbosity
+	subprocess.check_call(args,
+			env=environ, cwd=workdir,
+			stdout=output, stderr=output)
 
-	run_tool(cfg, args, tool_pkgs=tool_pkgs, virtual_tools=virtual_tools, workdir=workdir,
-			extra_environ=environ, for_package=for_package,
-			quiet=quiet)
+def run_program(cfg, context, subject, args,
+		tool_pkgs=[], virtual_tools=[], workdir=None, extra_environ=dict(),
+		for_package=False, containerless=False, quiet=False):
+	pkg_queue = []
+	pkg_visited = set()
+
+	for pkg in tool_pkgs:
+		assert pkg.name not in pkg_visited
+		pkg_queue.append(pkg)
+		pkg_visited.add(pkg.name)
+
+	i = 0 # Need index-based loop as pkg_queue is mutated in the loop.
+	while i < len(pkg_queue):
+		pkg = pkg_queue[i]
+		for dep_name in pkg.recursive_tools_required:
+			if dep_name in pkg_visited:
+				continue
+			dep_pkg = cfg.get_tool_pkg(dep_name)
+			pkg_queue.append(dep_pkg)
+			pkg_visited.add(dep_name)
+		i += 1
+
+	manifest = {
+		'context': context,
+		'args': args,
+		'workdir': workdir,
+		'extra_environ': extra_environ,
+		'quiet': quiet,
+		'for_package': for_package,
+		'virtual_tools': list(virtual_tools),
+		'tools': [],
+		'sysroot_subdir': cfg.sysroot_subdir,
+		'option_values': {name: cfg.get_option_value(name) for name in cfg.all_options}
+	}
+
+	if context == 'source':
+		manifest['subject'] = {
+			'source_subdir': subject.source_subdir
+		}
+	elif context == 'tool':
+		src = cfg.get_source(subject.source)
+		manifest['subject'] = {
+			'source_subdir': src.source_subdir,
+			'build_subdir': subject.build_subdir,
+			'prefix_subdir': subject.prefix_subdir
+		}
+	elif context == 'tool-stage':
+		tool = subject.pkg
+		src = cfg.get_source(tool.source)
+		manifest['subject'] = {
+			'source_subdir': src.source_subdir,
+			'build_subdir': tool.build_subdir,
+			'prefix_subdir': tool.prefix_subdir
+		}
+	elif context == 'pkg':
+		src = cfg.get_source(subject.source)
+		manifest['subject'] = {
+			'source_subdir': src.source_subdir,
+			'build_subdir': subject.build_subdir,
+			'collect_subdir': subject.collect_subdir
+		}
+	elif context == 'tool-task':
+		tool = subject.pkg
+		src = cfg.get_source(tool.source)
+		manifest['subject'] = {
+			'source_subdir': src.source_subdir,
+			'build_subdir': tool.build_subdir,
+			'prefix_subdir': tool.prefix_subdir
+		}
+	elif context == 'pkg-task':
+		pkg = subject.pkg
+		src = cfg.get_source(pkg.source)
+		manifest['subject'] = {
+			'source_subdir': src.source_srcdir,
+			'build_subdir': pkg.build_subdir,
+			'collect_subdir': pkg.collect_subdir
+		}
+
+	for tool in pkg_queue:
+		manifest['tools'].append({
+			'prefix_subdir': tool.prefix_subdir,
+			'exports_shared_libs': tool.exports_shared_libs,
+			'exports_aclocal': tool.exports_aclocal
+		})
+
+	runtime = cfg.container_runtime
+	container_yml = cfg._site_yml.get('container', dict())
+
+	use_container = True
+	if containerless:
+		if container_yml.get('allow_containerless', False):
+			use_container = False
+	if runtime is None:
+		use_container = False
+
+	if use_container:
+		if runtime == 'dummy':
+			manifest['source_root'] = cfg.source_root
+			manifest['build_root'] = cfg.build_root
+
+			print("{}xbstrap{}: Running {} (tools: {}) in dummy container".format(
+					colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+					args, [tool.name for tool in pkg_queue]))
+
+			if debug_manifests:
+				print(yaml.dump(manifest))
+
+			proc = subprocess.Popen(['xbstrap', 'execute-manifest',
+					'-c', yaml.dump(manifest)])
+			proc.wait()
+			if proc.returncode != 0:
+				raise ProgramFailureException()
+		else:
+			assert runtime == 'docker'
+			if any(prop not in container_yml for prop in ['src_mount', 'build_mount', 'image']):
+				raise RuntimeError("Docker runtime requires src_mount, build_mount and image properties")
+
+
+			manifest['source_root'] = container_yml['src_mount']
+			manifest['build_root'] = container_yml['build_mount']
+
+			print("{}xbstrap{}: Running {} (tools: {}) in Docker".format(
+					colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+					args, [tool.name for tool in pkg_queue]))
+
+			if debug_manifests:
+				print(yaml.dump(manifest))
+
+			docker_args = ['docker', 'run', '--rm', '-i', '--init',
+				'-v', cfg.source_root + ':' + container_yml['src_mount'],
+				'-v', cfg.build_root + ':' + container_yml['build_mount']]
+			if os.isatty(0): # FD zero = stdin.
+				docker_args += ['-t']
+			if 'create_extra_args' in container_yml:
+				docker_args += container_yml['create_extra_args']
+			docker_args += [container_yml['image'],
+				'xbstrap', 'execute-manifest', '-c', yaml.dump(manifest)]
+			proc = subprocess.Popen(docker_args)
+			proc.wait()
+			if proc.returncode != 0:
+				raise ProgramFailureException()
+	else:
+		manifest['source_root'] = cfg.source_root
+		manifest['build_root'] = cfg.build_root
+
+		print("{}xbstrap{}: Running {} (tools: {})".format(
+				colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+				args, [tool.name for tool in pkg_queue]))
+
+		if debug_manifests:
+			print(yaml.dump(manifest))
+
+		execute_manifest(manifest)
+
+def run_step(cfg, context, subject, step, tool_pkgs, virtual_tools,
+		for_package=False):
+	run_program(cfg, context, subject, step.args,
+			tool_pkgs=tool_pkgs, virtual_tools=virtual_tools, workdir=step.workdir,
+			extra_environ=step.environ,
+			for_package=for_package,
+			containerless=step.containerless,
+			quiet=step.quiet and not verbosity)
 
 def postprocess_libtool(cfg, pkg):
 	for libdir in ['lib', 'lib64', 'lib32', 'usr/lib', 'usr/lib64', 'usr/lib32']:
@@ -1598,19 +1836,7 @@ def regenerate_src(cfg, src):
 		for dep_name in src.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
-		def substitute(varname):
-			if varname == 'SOURCE_ROOT':
-				return cfg.source_root
-			elif varname == 'BUILD_ROOT':
-				return cfg.build_root
-			elif varname == 'SYSROOT_DIR':
-				return cfg.sysroot_dir
-			elif varname == 'THIS_SOURCE_DIR':
-				return src.source_dir
-			elif varname.startswith('OPTION:'):
-				return cfg.get_option_value(varname[7:])
-
-		run_step(cfg, step, src.source_dir, substitute, tool_pkgs, src.virtual_tools)
+		run_step(cfg, 'source', src, step, tool_pkgs, src.virtual_tools)
 
 	src.mark_as_regenerated()
 
@@ -1629,26 +1855,7 @@ def configure_tool(cfg, pkg):
 		for dep_name in pkg.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
-		def substitute(varname):
-			if varname == 'SOURCE_ROOT':
-				return cfg.source_root
-			elif varname == 'BUILD_ROOT':
-				return cfg.build_root
-			elif varname == 'SYSROOT_DIR':
-				return cfg.sysroot_dir
-			elif varname == 'THIS_SOURCE_DIR':
-				return src.source_dir
-			elif varname == 'THIS_BUILD_DIR':
-				return pkg.build_dir
-			elif varname == 'PREFIX':
-				return pkg.prefix_dir
-			elif varname == 'PARALLELISM':
-				nthreads = get_concurrency()
-				return str(nthreads)
-			elif varname.startswith('OPTION:'):
-				return cfg.get_option_value(varname[7:])
-
-		run_step(cfg, step, pkg.build_dir, substitute, tool_pkgs, pkg.virtual_tools)
+		run_step(cfg, 'tool', pkg, step, tool_pkgs, pkg.virtual_tools)
 
 	pkg.mark_as_configured()
 
@@ -1661,26 +1868,7 @@ def compile_tool_stage(cfg, stage):
 		for dep_name in pkg.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
-		def substitute(varname):
-			if varname == 'SOURCE_ROOT':
-				return cfg.source_root
-			elif varname == 'BUILD_ROOT':
-				return cfg.build_root
-			elif varname == 'SYSROOT_DIR':
-				return cfg.sysroot_dir
-			elif varname == 'THIS_SOURCE_DIR':
-				return src.source_dir
-			elif varname == 'THIS_BUILD_DIR':
-				return pkg.build_dir
-			elif varname == 'PREFIX':
-				return pkg.prefix_dir
-			elif varname == 'PARALLELISM':
-				nthreads = get_concurrency()
-				return str(nthreads)
-			elif varname.startswith('OPTION:'):
-				return cfg.get_option_value(varname[7:])
-
-		run_step(cfg, step, pkg.build_dir, substitute, tool_pkgs, pkg.virtual_tools)
+		run_step(cfg, 'tool-stage', stage, step, tool_pkgs, pkg.virtual_tools)
 
 	stage.mark_as_compiled()
 
@@ -1717,23 +1905,7 @@ def install_tool_stage(cfg, stage):
 		for dep_name in tool.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
-		def substitute(varname):
-			if varname == 'SOURCE_ROOT':
-				return cfg.source_root
-			elif varname == 'BUILD_ROOT':
-				return cfg.build_root
-			elif varname == 'SYSROOT_DIR':
-				return cfg.sysroot_dir
-			elif varname == 'THIS_SOURCE_DIR':
-				return src.source_dir
-			elif varname == 'THIS_BUILD_DIR':
-				return tool.build_dir
-			elif varname == 'PREFIX':
-				return tool.prefix_dir
-			elif varname.startswith('OPTION:'):
-				return cfg.get_option_value(varname[7:])
-
-		run_step(cfg, step, tool.build_dir, substitute, tool_pkgs, tool.virtual_tools)
+		run_step(cfg, 'tool-stage', stage, step, tool_pkgs, tool.virtual_tools)
 
 	stage.mark_as_installed()
 
@@ -1757,26 +1929,7 @@ def configure_pkg(cfg, pkg):
 		for dep_name in pkg.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
-		def substitute(varname):
-			if varname == 'SOURCE_ROOT':
-				return cfg.source_root
-			elif varname == 'BUILD_ROOT':
-				return cfg.build_root
-			elif varname == 'SYSROOT_DIR':
-				return cfg.sysroot_dir
-			elif varname == 'THIS_SOURCE_DIR':
-				return src.source_dir
-			elif varname == 'THIS_BUILD_DIR':
-				return pkg.build_dir
-			elif varname == 'PARALLELISM':
-				nthreads = get_concurrency()
-				return str(nthreads)
-			elif varname.startswith('OPTION:'):
-				return cfg.get_option_value(varname[7:])
-			elif varname.startswith('OPTION:'):
-				return cfg.get_option_value(varname[7:])
-
-		run_step(cfg, step, pkg.build_dir, substitute, tool_pkgs, pkg.virtual_tools,
+		run_step(cfg, 'pkg', pkg, step, tool_pkgs, pkg.virtual_tools,
 				for_package=True)
 
 	pkg.mark_as_configured()
@@ -1793,26 +1946,7 @@ def build_pkg(cfg, pkg, reproduce=False):
 		for dep_name in pkg.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
-		def substitute(varname):
-			if varname == 'SOURCE_ROOT':
-				return cfg.source_root
-			elif varname == 'BUILD_ROOT':
-				return cfg.build_root
-			elif varname == 'SYSROOT_DIR':
-				return cfg.sysroot_dir
-			elif varname == 'THIS_SOURCE_DIR':
-				return src.source_dir
-			elif varname == 'THIS_BUILD_DIR':
-				return pkg.build_dir
-			elif varname == 'THIS_COLLECT_DIR':
-				return pkg.collect_dir
-			elif varname == 'PARALLELISM':
-				nthreads = get_concurrency()
-				return str(nthreads)
-			elif varname.startswith('OPTION:'):
-				return cfg.get_option_value(varname[7:])
-
-		run_step(cfg, step, pkg.build_dir, substitute, tool_pkgs, pkg.virtual_tools,
+		run_step(cfg, 'pkg', pkg, step, tool_pkgs, pkg.virtual_tools,
 				for_package=True)
 
 	postprocess_libtool(cfg, pkg)
@@ -1975,7 +2109,7 @@ def pull_pkg_pack(cfg, pkg):
 	print('{}xbstrap{}: Downloading x86_64-repodata from {}'.format(
 			colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
 			repo_url))
-	util.interactive_download(rd_url, rd_path)
+	_util.interactive_download(rd_url, rd_path)
 
 	# Find the package within the repodata's index file.
 	index = _xbps_utils.read_repodata(rd_path)
@@ -1989,7 +2123,7 @@ def pull_pkg_pack(cfg, pkg):
 	print('{}xbstrap{}: Downloading {} from {}'.format(
 			colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
 			xbps_file, repo_url))
-	util.interactive_download(pkg_url, os.path.join(cfg.xbps_repository_dir, xbps_file))
+	_util.interactive_download(pkg_url, os.path.join(cfg.xbps_repository_dir, xbps_file))
 
 	# Run xbps-rindex.
 	output = subprocess.DEVNULL
@@ -2009,20 +2143,7 @@ def run_task(cfg, task):
 	for dep_name in task.tool_dependencies:
 		tools_required.append(cfg.get_tool_pkg(dep_name))
 
-	def substitute(varname):
-		if varname == 'SOURCE_ROOT':
-			return cfg.source_root
-		elif varname == 'BUILD_ROOT':
-			return cfg.build_root
-		elif varname == 'SYSROOT_DIR':
-			return cfg.sysroot_dir
-		elif varname == 'PARALLELISM':
-			nthreads = get_concurrency()
-			return str(nthreads)
-		elif varname.startswith('OPTION:'):
-			return cfg.get_option_value(varname[7:])
-
-	run_step(cfg, task.script_step, cfg.source_root, substitute, tools_required, task.virtual_tools,
+	run_step(cfg, 'task', task, task.script_step, tools_required, task.virtual_tools,
 			for_package=False)
 
 
@@ -2033,26 +2154,7 @@ def run_pkg_task(cfg, task):
 	for dep_name in task.pkg.tool_dependencies:
 		tools_required.append(cfg.get_tool_pkg(dep_name))
 
-	def substitute(varname):
-		if varname == 'SOURCE_ROOT':
-			return cfg.source_root
-		elif varname == 'BUILD_ROOT':
-			return cfg.build_root
-		elif varname == 'SYSROOT_DIR':
-			return cfg.sysroot_dir
-		elif varname == 'THIS_SOURCE_DIR':
-			return src.source_dir
-		elif varname == 'THIS_BUILD_DIR':
-			return task._pkg.build_dir
-		elif varname == 'PREFIX':
-			return task._pkg.prefix_dir
-		elif varname == 'PARALLELISM':
-			nthreads = get_concurrency()
-			return str(nthreads)
-		elif varname.startswith('OPTION:'):
-			return cfg.get_option_value(varname[7:])
-
-	run_step(cfg, task.script_step, task.pkg.build_dir, substitute, tools_required,
+	run_step(cfg, 'pkg-task', task, task.script_step, tools_required,
 			task.pkg.virtual_tools, for_package=False)
 
 def run_tool_task(cfg, task):
@@ -2062,26 +2164,7 @@ def run_tool_task(cfg, task):
 	for dep_name in task.pkg.tool_dependencies:
 		tools_required.append(cfg.get_tool_pkg(dep_name))
 
-	def substitute(varname):
-		if varname == 'SOURCE_ROOT':
-			return cfg.source_root
-		elif varname == 'BUILD_ROOT':
-			return cfg.build_root
-		elif varname == 'SYSROOT_DIR':
-			return cfg.sysroot_dir
-		elif varname == 'THIS_SOURCE_DIR':
-			return src.source_dir
-		elif varname == 'THIS_BUILD_DIR':
-			return task._pkg.build_dir
-		elif varname == 'PREFIX':
-			return task._pkg.prefix_dir
-		elif varname == 'PARALLELISM':
-			nthreads = get_concurrency()
-			return str(nthreads)
-		elif varname.startswith('OPTION:'):
-			return cfg.get_option_value(varname[7:])
-
-	run_step(cfg, task.script_step, task.pkg.build_dir, substitute, tools_required,
+	run_step(cfg, 'tool-task', task, task.script_step, tools_required,
 			task.pkg.virtual_tools, for_package=False)
 
 # ---------------------------------------------------------------------------------------
@@ -2215,6 +2298,10 @@ class PlanItem:
 		}
 		self._state = visitors[self.action](self.subject, self.settings)
 
+class ProgramFailureException(Exception):
+	def __init__(self):
+		super().__init__("Program failed")
+
 class ExecutionFailureException(Exception):
 	def __init__(self, step, subject):
 		super().__init__("Action {} of {} {} failed".format(Action.strings[step],
@@ -2712,7 +2799,7 @@ class Plan:
 					raise AssertionError("Unexpected action")
 				item.exec_status = ExecutionStatus.SUCCESS
 				emit_progress('success')
-			except (subprocess.CalledProcessError, ExecutionFailureException):
+			except (subprocess.CalledProcessError, ProgramFailureException, ExecutionFailureException):
 				item.exec_status = ExecutionStatus.STEP_FAILED
 				emit_progress('failure')
 				if not self.keep_going:
