@@ -1,7 +1,18 @@
+# SPDX-License-Identifier: MIT
+
 import os
 import urllib.parse
 import urllib.request
 
+def build_environ_paths(environ, varname, prepend):
+	if not prepend:
+		return
+	joined = ':'.join(prepend)
+	if varname in environ and environ[varname]:
+		environ[varname] = joined + ':' + environ[varname]
+	else:
+		environ[varname] = joined
+
 def interactive_download(url, path):
 	print('...', end='') # This will become the status line.
 
